/*
Copyright (C) 1994-2000 The XFree86 Project, Inc.  All Rights Reserved.
Copyright (C) 2000 Silicon Motion, Inc.  All Rights Reserved.
Copyright (C) 2004 Himalaya Software.  All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
XFREE86 PROJECT BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the names of the XFree86 Project and
Silicon Motion shall not be used in advertising or otherwise to promote the
sale, use or other dealings in this Software without prior written
authorization from the XFree86 Project and Silicon Motion.
*/
/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/siliconmotion/smi_shadow.c,v 1.2 2000/12/05 21:18:37 dawes Exp $ */

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86Resources.h"
#include "xf86_ansic.h"
#include "xf86PciInfo.h"
#include "xf86Pci.h"
#include "shadowfb.h"
#include "servermd.h"

#include "smi.h"

static void SMI_RotateRect(ScrnInfoPtr pScrn, int x1, int y1, int x2, int y2,
                           int w, int h);

/******************************************************************************\
|* SMI_RefreshArea
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  num     Number of boxes to refresh.
|*                  pbox    Pointer to an array of boxes to refresh.
|*
|*  DESCRIPTION:    Refresh a portion of the shadow buffer to the visual screen
|*                  buffer.  This is mainly used for rotation purposes.
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void SMI_RefreshArea(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int width, height, srcX, srcY;

    ENTER_PROC("SMI_RefreshArea");

    /* #671 */
    if (pSmi->polyLines)
    {
        pSmi->polyLines = FALSE;
        LEAVE_PROC("SMI_RefreshArea");
        return;
    }

    /* Ignore update request if already updated by QuickRotate. */
    if (pSmi->ignoreRequest)
    {
        pSmi->ignoreRequest = FALSE;
        LEAVE_PROC("SMI_RefreshArea");
        return;
    }

    /* Setup the hardware here. */
    WaitQueue(1);
    DEST_BASE(pSmi->FBOffset >> 3);

    /* #672 */
    if (pSmi->ClipTurnedOn)
    {
        WaitQueue(1);
        WRITE_DPR(pSmi, 0x2C, pSmi->ScissorsLeft);
        pSmi->ClipTurnedOn = FALSE;
    }

    while (num--)
    {
        /* Get coordinates of the box to refresh. */
        srcX   = pbox->x1;
        srcY   = pbox->y1;
        width  = pbox->x2 - srcX;
        height = pbox->y2 - srcY;

        if ((width > 0) && (height > 0))
            SMI_RotateRect(pScrn, srcX, srcY, srcX, srcY, width, height);

        pbox++;
    }

    /* Restore the hardware here. */
    WaitQueue(1);
    DEST_BASE(0);

    LEAVE_PROC("SMI_RefreshArea");
}

/******************************************************************************\
|* SMI_PointerMoved
|*=============================================================================
|*
|*  PARAMETERS:     index   Index of current screen.
|*                  x       X location of pointer.
|*                  y       Y location of pointer.
|*
|*  DESCRIPTION:    Adjust the pointer location if we are in rotation mode.
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void SMI_PointerMoved(int index, int x, int y)
{
    ScrnInfoPtr pScrn = xf86Screens[index];
    SMIPtr pSmi = SMIPTR(pScrn);
    int newX, newY;

    switch (pSmi->rotate)
    {
    case SMI_ROTATE_CW:
        /* 90 degrees CW rotation. */
        newX = pSmi->ShadowWidth - y - 1;
        newY = x;
        break;

    case SMI_ROTATE_CCW:
        /* 90 degrees CCW rotation. */
        newX = y;
        newY = pSmi->ShadowHeight - x - 1;
        break;

    default:
        /* No rotation. */
        newX = x;
        newY = y;
        break;
    }

    /* Pass adjusted pointer coordinates original PointerMoved function. */
    (*pSmi->PointerMoved)(index, newX, newY);
}

/******************************************************************************\
|* SMI_UpdateSolidFill
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  x       X coordinate of rectangle to fill.
|*                  y       Y coordinate of rectangle to fill.
|*                  w       Width of rectangle to fill.
|*                  h       Height of rectangle to fill.
|*
|*  DESCRIPTION:    Fill a rectangle in the shadow buffer with a preloaded solid
|*                  color (setup by the acceleration code).
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void
SMI_UpdateSolidFill(ScrnInfoPtr pScrn, int x, int y, int w, int h)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int destX, destY, width, height;

    ENTER_PROC("SMI_UpdateSolidFill");

    /* Setup the hardware here. */
    WaitQueue(1);
    DEST_BASE(pSmi->FBOffset >> 3);

    /* #672 */
    if (pSmi->ClipTurnedOn)
    {
        WaitQueue(1);
        WRITE_DPR(pSmi, 0x2C, pSmi->ScissorsLeft);
        pSmi->ClipTurnedOn = FALSE;
    }

    switch (pSmi->rotate)
    {
    case SMI_ROTATE_CW:
        destX = pSmi->ShadowWidth - y - h;
        destY = x;
        width = h;
        height = w;
        break;

    case SMI_ROTATE_CCW:
        destX = y;
        destY = pSmi->ShadowHeight - x - w;
        width = h;
        height = w;
        break;

    default:
        destX = x;
        destY = y;
        width = w;
        height = h;
        break;
    }

    WaitQueue(3);
    DEST_XY(destX, destY);
    SIZE_XY(width, height);
    START(pSmi->AccelCmd);

    /* Restore the hardware here. */
    WaitQueue(1);
    DEST_BASE(0);

    LEAVE_PROC("SMI_UpdateSolidFill");
}

/******************************************************************************\
|* SMI_UpdateCopy
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  x1      X coordinate of source rectangle to copy.
|*                  y1      Y coordinate of source rectangle to copy.
|*                  x2      X coordinate of destination rectangle to copy.
|*                  y2      Y coordinate of destination rectangle to copy.
|*                  w       Width of rectangle to copy.
|*                  h       Height of rectangle to copy.
|*
|*  DESCRIPTION:    Copy a rectangle in the shadow buffer (setup by the
|*                  acceleration code).
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void
SMI_UpdateCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2, int y2, int w, int h)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int srcX, srcY, destX, destY, width, height;
    CARD32 accelCmd = pSmi->AccelCmd;

    ENTER_PROC("SMI_UpdateCopy");

    /* Setup the hardware here. */
    WaitQueue(2);
    SRC_BASE(pSmi->FBOffset >> 3);
    DEST_BASE(pSmi->FBOffset >> 3);

    /* #672 */
    if (pSmi->ClipTurnedOn)
    {
        WaitQueue(1);
        WRITE_DPR(pSmi, 0x2C, pSmi->ScissorsLeft);
        pSmi->ClipTurnedOn = FALSE;
    }

    switch (pSmi->rotate)
    {
    case SMI_ROTATE_CW:
        srcX = pSmi->ShadowWidth - y1 - h;
        srcY = x1;
        destX = pSmi->ShadowWidth - y2 - h;
        destY = x2;
        width = h;
        height = w;
        break;

    case SMI_ROTATE_CCW:
        srcX = y1;
        srcY = pSmi->ShadowHeight - x1 - w;
        destX = y2;
        destY = pSmi->ShadowHeight - x2 - w;
        width = h;
        height = w;
        break;

    default:
        srcX = x1;
        srcY = y1;
        destX = x2;
        destY = y2;
        width = w;
        height = h;
        break;
    }

    if ((srcY == destY && srcX < destX) || srcY < destY)
    {
        srcX += width - 1;
        srcY += height - 1;
        destX += width - 1;
        destY += height - 1;
        accelCmd |= SMI_RIGHT_TO_LEFT;
    }
    else
        accelCmd &= ~SMI_RIGHT_TO_LEFT;

    WaitQueue(4);
    SRC_XY(srcX, srcY);
    DEST_XY(destX, destY);
    SIZE_XY(width, height);
    START(accelCmd);

    /* Restore the hardware here. */
    WaitQueue(2);
    SRC_BASE(0);
    DEST_BASE(0);

    LEAVE_PROC("SMI_UpdateCopy");
}

/******************************************************************************\
|* SMI_UpdateRotate
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  x1      X coordinate of source rectangle to copy.
|*                  y1      Y coordinate of source rectangle to copy.
|*                  x2      X coordinate of destination rectangle to copy.
|*                  y2      Y coordinate of destination rectangle to copy.
|*                  w       Width of rectangle to copy.
|*                  h       Height of rectangle to copy.
|*
|*  DESCRIPTION:    Copy a rectangle from a pixmap to the shadow buffer (setup
|*                  by the acceleration code).
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void
SMI_UpdateRotate(ScrnInfoPtr pScrn, int x1, int y1, int x2, int y2, int w,
                 int h)
{
    SMIPtr pSmi = SMIPTR(pScrn);

    ENTER_PROC("SMI_UpdateRotate");

    /* Setup the hardware here. */
    WaitQueue(1);
    DEST_BASE(pSmi->FBOffset >> 3);

    /* #672 */
    if (pSmi->ClipTurnedOn)
    {
        WaitQueue(1);
        WRITE_DPR(pSmi, 0x2C, pSmi->ScissorsLeft);
        pSmi->ClipTurnedOn = FALSE;
    }

    SMI_RotateRect(pScrn, x1, y1, x2, y2, w, h);

    /* Restore the hardware here. */
    WaitQueue(1);
    DEST_BASE(0);

    LEAVE_PROC("SMI_UpdateRotate");
}

Bool
SMI_RotatePattern(ScrnInfoPtr pScrn, int patx, int paty)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int x, y;
    unsigned int pat, bit;
    unsigned char pattern[8];

    ENTER_PROC("SMI_RotatePattern");

    memset(pattern, 0, 8);

    switch (pSmi->rotate)
    {
    case SMI_ROTATE_CW:
        for (y = 0; y < 8; y++)
        {
            if (y < 4)
                pat = patx >> (y * 8);
            else
                pat = paty >> ((y - 4) * 8);

            for (x = 0; x < 8; x++)
            {
                bit = (pat >> (7 - x)) & 1;
                pattern[7 - x] |= bit << (7 - y);
            }
        }
        break;

    case SMI_ROTATE_CCW:
        for (y = 0; y < 8; y++)
        {
            if (y < 4)
                pat = patx >> (y * 8);
            else
                pat = paty >> ((y - 4) * 8);

            for (x = 0; x < 8; x++)
            {
                bit = (pat >> (7 - x)) & 1;
                pattern[x] |= bit << y;
            }
        }
        break;

    default:
        LEAVE_PROC("SMI_RotatePattern - FALSE");
        return(FALSE);
    }

    WaitQueue(2);
    PATTERN(*(unsigned int*) &pattern[0], *(unsigned int*) &pattern[4]);

    LEAVE_PROC("SMI_RotatePattern - TRUE");
    return(TRUE);
}

static void
SMI_RotateRect(ScrnInfoPtr pScrn, int x1, int y1, int x2, int y2, int w, int h)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int destX, destY;
    int maxWidth, stripe;

    ENTER_PROC("SMI_RotateRect");
    DEBUG((VERBLEV, "src=%d,%d dest=%d,%d size=%dx%d\n", x1, y1, x2, y2, w, h));

    /* SM731 cannot rotate-blt more than a certain number of pixels. */
    if (pSmi->Chipset == SMI_COUGAR3DR)
        maxWidth = 1280 / pScrn->bitsPerPixel;
    else
        maxWidth = 4096;

    switch (pSmi->rotate)
    {
    case SMI_ROTATE_CW:
        /* 90 degrees CW rotation.  Rotate destination coordinates:

            *---+
            |   |       +-----*
            |   |       |     |  destX = ShadowWidth - y2 - 1
            |   |  -->  |     |  destY = x2
            |   |       |     |
            |   |       +-----+
            +---+
        */
        destX = pSmi->ShadowWidth - y2 - 1;
        destY = x2;

        for (; w > 0; w -= stripe)
        {
            stripe = min(w, maxWidth);

            WaitQueue(4);
            SRC_XY(x1, y1);
            DEST_XY(destX, destY);
            SIZE_XY(stripe, h);
            START(0xCC | SMI_ROTATE_BLT | SMI_ROTATE_CW | SMI_START_ENGINE);
            destY += stripe;
            x1 += stripe;
        }
        break;

    case SMI_ROTATE_CCW:
        /* 90 degrees CCW rotatation.  Rotate destination coordinates:

            *---+
            |   |       +-----+
            |   |       |     |  destX = y2
            |   |  -->  |     |  destY = ShadowHeight - x2 - 1
            |   |       |     |
            |   |       *-----+
            +---+
        */
        destX = y2;
        destY = pSmi->ShadowHeight - x2 - 1;

        for (; w > 0; w -= stripe)
        {
            stripe = min(w, maxWidth);

            WaitQueue(4);
            SRC_XY(x1, y1);
            DEST_XY(destX, destY);
            SIZE_XY(stripe, h);
            START(0xCC | SMI_ROTATE_BLT | SMI_ROTATE_CCW | SMI_START_ENGINE);
            destY -= stripe;
            x1 += stripe;
        }
        break;

    default:
        /* No rotation, perform a normal copy. */
        if (pScrn->bitsPerPixel == 24)
        {
            x1 *= 3;
            x2 *= 3;
            w  *= 3;

            if (pSmi->Chipset == SMI_LYNX)
            {
                y1 *= 3;
                y2 *= 3;
            }
        }

        WaitQueue(4);
        SRC_XY(x1, y1);
        DEST_XY(x2, y2);
        SIZE_XY(w, h);
        START(0xCC | SMI_BITBLT | SMI_START_ENGINE);
        break;
    }

    LEAVE_PROC("SMI_RotateRect");
}

/******************************************************************************\
|* SMI_RefreshArea8
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  num     Number of boxes to refresh.
|*                  pbox    Pointer to an array of boxes to refresh.
|*
|*  DESCRIPTION:    Software algorithm for refreshing a portion of the 8-bpp
|*                  shadow buffer to the visual screen buffer.  This is used for
|*                  rotation purposes.
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void
SMI_RefreshArea8(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int count, width, height, y1, y2, dstPitch, srcPitch;
    CARD8* dstPtr;
    CARD8* srcPtr;
    CARD8* src;
    CARD32* dst;

    ENTER_PROC("SMI_RefreshArea8");

    srcPitch = pSmi->fbWidth;
    dstPitch = pSmi->memWidth;

    if (pSmi->rotate == SMI_ROTATE_CW)
    {
        while (num--)
        {
            width = pbox->x2 - pbox->x1;
            y1 = pbox->y1 & ~3;
            y2 = (pbox->y2 + 3) & ~3;
            height = (y2 - y1) >> 2;

            srcPtr = (CARD8*) pSmi->fbPointer + pbox->x1 +
                     ((y2 - 1) * srcPitch);
            dstPtr = (CARD8*) pSmi->FBBase + (pSmi->ShadowWidth - y2) +
                     (pbox->x1 * dstPitch);

            while (width--)
            {
                src = srcPtr;
                dst = (CARD32*) dstPtr;
                count = height;
                while (count--)
                {
                    *(dst++) = src[0] | (src[-srcPitch] << 8) | 
                               (src[-srcPitch * 2] << 16) |
                               (src[-srcPitch * 3] << 24);
                    src -= srcPitch * 4;
                }
                srcPtr += 1;
                dstPtr += dstPitch;
            }

            pbox++;
        }
    }
    else
    {
        while (num--)
        {
            width = pbox->x2 - pbox->x1;
            y1 = pbox->y1 & ~1;
            y2 = (pbox->y2 + 1) & ~1;
            height = (y2 - y1) >> 1;

            srcPtr = (CARD8*) pSmi->fbPointer + (pbox->x2 - 1) +
                     (y1 * srcPitch);
            dstPtr = (CARD8*) pSmi->FBBase + y1 +
                     ((pSmi->ShadowHeight - pbox->x2) * dstPitch);

            while (width--)
            {
                src = srcPtr;
                dst = (CARD32*) dstPtr;
                count = height;
                while (count--)
                {
                    *(dst++) = src[0] | (src[srcPitch] << 8) |
                               (src[srcPitch * 2] << 16) |
                               (src[srcPitch * 3] << 24);
                    src += srcPitch * 4;
                }
                srcPtr -= 1;
                dstPtr += dstPitch;
            }

            pbox++;
        }
    }

    LEAVE_PROC("SMI_RefreshArea8");
}

/******************************************************************************\
|* SMI_RefreshArea16
|*=============================================================================
|*
|*  PARAMETERS:     pScrn   Pointer to ScrnInfo structure.
|*                  num     Number of boxes to refresh.
|*                  pbox    Pointer to an array of boxes to refresh.
|*
|*  DESCRIPTION:    Software algorithm for refreshing a portion of the 16-bpp
|*                  shadow buffer to the visual screen buffer.  This is used for
|*                  rotation purposes.
|*
|*  RETURNS:        Nothing.
|*
\******************************************************************************/
void
SMI_RefreshArea16(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    SMIPtr pSmi = SMIPTR(pScrn);
    int count, width, height, y1, y2, dstPitch, srcPitch;
    CARD16* dstPtr;
    CARD16* srcPtr;
    CARD16* src;
    CARD32* dst;

    ENTER_PROC("SMI_RefreshArea16");

    srcPitch = pSmi->fbWidth;
    dstPitch = pSmi->memWidth;

    if (pSmi->rotate == SMI_ROTATE_CW)
    {
        while (num--)
        {
            width = pbox->x2 - pbox->x1;
            y1 = pbox->y1 & ~1;
            y2 = (pbox->y2 + 1) & ~1;
            height = (y2 - y1) >> 1;

            srcPtr = (CARD16*) pSmi->fbPointer + pbox->x1 +
                     ((y2 - 1) * srcPitch);
            dstPtr = (CARD16*) pSmi->FBBase + (pSmi->ShadowWidth - y2) +
                     (pbox->x1 * dstPitch);

            while (width--)
            {
                src = srcPtr;
                dst = (CARD32*) dstPtr;
                count = height;
                while (count--)
                {
                    *(dst++) = src[0] | (src[-srcPitch] << 16);
                    src -= srcPitch * 2;
                }
                srcPtr += 1;
                dstPtr += dstPitch;
            }

            pbox++;
        }
    }
    else
    {
        while (num--)
        {
            width = pbox->x2 - pbox->x1;
            y1 = pbox->y1 & ~1;
            y2 = (pbox->y2 + 1) & ~1;
            height = (y2 - y1) >> 1;

            srcPtr = (CARD16*) pSmi->fbPointer + (pbox->x2 - 1) +
                     (y1 * srcPitch);
            dstPtr = (CARD16*) pSmi->FBBase + y1 +
                     ((pSmi->ShadowHeight - pbox->x2) * dstPitch);

            while (width--)
            {
                src = srcPtr;
                dst = (CARD32*) dstPtr;
                count = height;
                while (count--)
                {
                    *(dst++) = src[0] | (src[srcPitch] << 16);
                    src += srcPitch * 2;
                }
                srcPtr -= 1;
                dstPtr += dstPitch;
            }

            pbox++;
        }
    }

    LEAVE_PROC("SMI_RefreshArea16");
}
